package main.java.weka.classifiers.misc;

import static java.lang.System.out;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;
import libsvm.svm;
import libsvm.svm_model;
import libsvm.svm_node;
import libsvm.svm_parameter;
import libsvm.svm_problem;
import weka.classifiers.AbstractClassifier;
import weka.classifiers.functions.LinearRegression;
import weka.classifiers.functions.RBFRegressor;
import weka.core.Capabilities;
import weka.core.Capabilities.Capability;
import weka.core.matrix.Maths;
import weka.core.matrix.Matrix;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Utils;

public class ldsvm extends AbstractClassifier implements OptionHandler{
	
	//Parameters
	private int maxit;
	
	private double tol;
	
	private double beta;
	
	private int zLength;
	
	private Vector<AbstractClassifier> f = new Vector<AbstractClassifier>(zLength);
	
	private AbstractClassifier currentF;
	
	private svm_model g;
	
	private int regType;
	
	private String[] regOptions;
	
	public ldsvm(){
		this.maxit = 5; this.tol = .00001; this.beta = 1; this.zLength = 2; this.regType = 1;
		String [] optionStr = {""}; regOptions = optionStr;
	}
	
	public Capabilities getCapabilities(){
		Capabilities result = super.getCapabilities();
		result.disableAll();
		
		//Attributes
		result.enable(Capability.NUMERIC_ATTRIBUTES);
		
		//Classes
		result.enable(Capability.BINARY_CLASS);
		
		return result;
	}
	
	public double getBeta(){
		return beta;
	}
	
	public void setBeta(double newBeta){
		beta = newBeta;
	}
	
	public double getTol(){
		return tol;
	}
	
	public void setTol(double newTol){
		tol = newTol;
	}
	
	public int getMaxIt(){
		return maxit;
	}
	  
	public void setMaxIt(int newMaxIt) {
		maxit = newMaxIt;
	}
	
	public int getL(){
		return zLength;
	}
	
	public void setL(int newZLength){
		zLength = newZLength;
	}
	
	public int getRegType(){
		return regType;
	}
	
	public void setRegType(int newRegType){
		regType = newRegType;
		if(regType != 0 && regType != 1){
			regType = 0;
		}
	}
	
	public String getRegOptions(){
		StringBuilder build = new StringBuilder();
		for(int i = 0; i < regOptions.length; i++){
			build.append(regOptions[i]);
			build.append(" ");
		}
		return build.toString();
	}
	public void setRegOptions(String optionStr, int reg){
		StringTokenizer tok = new StringTokenizer(optionStr, " ");
		Vector<String> optionVector = new Vector<String>();
		if(optionStr.length() == 0){
			System.out.println("Options not processed, default will be used");
			
		}
		else{
			int temp = tok.countTokens();
			for(int i = 0; i < temp; i++){
				optionVector.add(tok.nextToken());
			}
		}
		if(reg == 1){
			Boolean N = false; Boolean C = false; Boolean capabilities = false;
			for(int i = 0; i < optionVector.size(); i++){
				switch(optionVector.get(i)){
					case "-N":
						N = true;
						break;
					case "-C":
						C = true;
						break;
					case "-do-not-check-capabilities":
						capabilities = true;
						break;
					default:
						break;
				}
			}
			if(N == false){
				optionVector.add("-N");
				optionVector.add("50");
			}
			if(C == false){
				optionVector.add("-C");
				optionVector.add("1");
			}
			if(capabilities == false){
				optionVector.add("-do-not-check-capabilities");
			}
		}
		if(reg == 0){
			Boolean S = false; Boolean C = false; Boolean minimal = false; Boolean R = false;
			Boolean capabilities = false;
			for(int i = 0; i < optionVector.size(); i++){
				switch(optionVector.get(i)){
					case "-S":
						S = true;
						break;
					case "-C":
						C = true;
						break;
					case "-minimal":
						minimal = true;
						break;
					case "-R":
						R = true;
						break;
					case "-do-not-check-capabilities":
						capabilities = true;
						break;
					default:
						break;
				}
			}
			if(S == false){
				optionVector.add("-S");
				optionVector.add("1");
			}
			if(C == false){
				optionVector.add("-C");
			}
			if(minimal == false){
				optionVector.add("-minimal");
			}
			if(R == false){
				optionVector.add("-R");
				optionVector.add("0");
			}
			if(capabilities == false){
				optionVector.add("-do-not-check-capabilities");
			}
		}
		regOptions = null;
		regOptions = new String[optionVector.size()];
		for(int i = 0; i < regOptions.length; i++){
			regOptions[i] = optionVector.get(i);
		}
	}
	public void buildClassifier(Instances dataX) throws Exception {
		
		svm.svm_set_print_string_function(new libsvm.svm_print_interface(){
	    @Override public void print(String s) {} // Disables svm output
	});		
		currentF = getRegressor(regType);
		currentF.setOptions(regOptions.clone());
		double[] E = new double[maxit+1];
		int N = dataX.numInstances();
		int D = dataX.numAttributes();
		int L = zLength;		
		Random rand = new Random();
		double[][] Z = new double[N][L];
		Matrix mZ = new Matrix(Z);
		Z = mZ.getArray();
		for(int i = 0; i < N; i++){
			for(int j = 0; j < L; j++){
				Z[i][j] = rand.nextDouble();
			}
		}
		//Normalization for Z
		for(int i = 0; i < L; i++){
			double tempNum = 0;
			for(int j = 0; j < N; j++){
				tempNum+=Z[j][i];
			}
			tempNum=tempNum/N;
			for(int j = 0; j < N; j++){
				Z[j][i]=Z[j][i]-tempNum;
			}
		}
		
		double tempNum = frob(mZ);
		for(int i = 0; i < N; i++){
			for(int j = 0; j < L; j++){
				Z[i][j] = Z[i][j]/tempNum;
			}
		}	
		//Setting Linear regression and linear svm options	
		svm_parameter param=new svm_parameter();
		param.svm_type=svm_parameter.C_SVC;	param.svm_type=svm_parameter.LINEAR;
		param.cache_size=20000; param.C=100; param.eps=0.0001;
		 
		//Formatting X
		double[] labels = dataX.attributeToDoubleArray(D-1);
		for(int i=0;i<labels.length;i++)
		{
			labels[i]= (2*labels[i])-1;		
		}
		Attribute zAtt = dataX.attribute(D-1);
		Attribute doubleAtt = new Attribute("dimReduction");
		dataX.replaceAttributeAt(doubleAtt, D-1);
		dataX.setClassIndex(D-1);		
		ArrayList<Attribute> attList = new ArrayList();
		for(int i = 0; i < L; i++){
			attList.add(new Attribute(Integer.toString(i)));
		}		
		attList.add(zAtt);
				
		//Format SVM problem
		svm_problem prob = new svm_problem();
		prob.x = new svm_node[N][]; prob.l = N; prob.y = new double[N];
		for(int i = 0; i < N; i++){
			prob.y[i]=labels[i];
			prob.x[i]=new svm_node[L];
			for(int j=0;j<L;j++){
				prob.x[i][j] = new svm_node();
				prob.x[i][j].index=j+1;	prob.x[i][j].value=Z[i][j];
			}
		}
		
		//Building f for the first time
		for(int i = 0; i < L; i++){
			for(int j = 0; j < N; j++){
				DenseInstance tempInst = (DenseInstance)dataX.get(j);
				tempInst.setValue(dataX.numAttributes()-1, Z[j][i]);
				dataX.set(j, tempInst);
			}
			currentF.buildClassifier(dataX);
			f.add(i, (AbstractClassifier) AbstractClassifier.makeCopy(currentF));
		}
		double[][] fX = fClassify(dataX, f);
		//Building g for the first time
		g = svm.svm_train(prob,param);									
		double[] gZ = gClassify(prob.x, g);
		double err = errClassification(labels, gZ);
		err= Math.floor(err*10000)/100;
		int l = g.sv_indices.length;
		System.out.print("Iteration #0, number of SVs=" + l + ", Classification error = " + err +"%");
		
		// Compute errors for the first time
		double  gErr = gError(Z,labels, g);
		double fErr = fError(Z, fX);
		E[0]= gErr+(beta*fErr);
		int newbeta=0;

		//Iterate,
		for(int m = 1; m <= maxit; m++){
			System.out.print("\rIteration #" + m);
			//Z-step, returns matrix Z
			Z = ZOpt(Z, fX, labels, beta, g);
			
			// Compute errors of steps
			gErr = gError(Z,labels, g);
			fErr = fError(Z, fX);
			E[m]= gErr+(beta*fErr);
			
			//F-step	
			f.removeAllElements();
			for(int i = 0; i < L; i++){
				for(int j = 0; j < N; j++){
					DenseInstance tempInst = (DenseInstance)dataX.get(j);
					tempInst.setValue(dataX.numAttributes()-1, Z[j][i]);
					dataX.set(j, tempInst);
				}
				currentF.buildClassifier(dataX);
				f.add(i, (AbstractClassifier)AbstractClassifier.makeCopy(currentF));
			}
			fX = fClassify(dataX, f);
			
			//G-step
			for(int i = 0; i < N; i++){
				for(int j=0;j<L;j++){					
					prob.x[i][j].value=Z[i][j];
				}
			}			
			g = svm.svm_train(prob,param);
			gZ = gClassify(prob.x, g);
			
			l = g.sv_indices.length;
			System.out.print(", number of SVs=" + l);
			err = errClassification(labels, gZ);
			err= Math.floor(err*10000)/100;
			System.out.print(" ,Classification error = " + err + "%");
			// fro1 = norm(Z,'fro'), fro2 = norm(Z-FX,'fro')
			double fro1 =  0, fro2 =0;
			for(int i = 0; i < N; i++)
			{
				for(int j=0;j<L;j++)
				{					
					fro1 += Math.pow(Z[i][j], 2);
					fro2 += Math.pow(Z[i][j]-fX[i][j], 2);
				}
			}		
			fro1 = Math.sqrt(fro1); fro2 = Math.sqrt(fro2);
			
			if(newbeta==1)
			{
				newbeta=0; //Update beta after at least 1 iteration.
			}
			else
			{				
				if(Math.abs(E[m]-E[m-1])<(tol*Math.abs(E[m-1])))
				{
					//Increase beta when convergence is achieved for fixed beta.
					beta=1.5*beta;newbeta=1;
					//Finish the program if constraints are satisfied.
					if(fro2 < tol*fro1)
					{
						break;
					}
				}
			}			
		}
		
		fX = fClassify(dataX, f);
		for(int i = 0; i < N; i++){
			for(int j=0;j<L;j++){					
				prob.x[i][j].value=fX[i][j];
			}
		}
		double[] gfX = new double[N];
		gfX = gClassify(prob.x, g);
		err = errClassification(labels, gfX);
		err= Math.floor(err*10000)/100;
		System.out.println("\rClassification error = " + err+"%");
	}
	public double errClassification(double[] labels, double[] pLabels)
	{
		int N=labels.length;
		int counter = 0;
		for(int i=0;i< N;i++)
		{
			if(labels[i]!=pLabels[i])
			{
				counter += 1;
			}
		}
		return (double) counter/N;
	}
	public double classifyInstance(Instance instance) throws Exception{
		double[] tempZ = new double[f.size()];
		for(int i = 0; i < tempZ.length; i++){
			tempZ[i] = f.get(i).classifyInstance(instance);
		}
		svm_node[][] node = new svm_node[1][tempZ.length];
		for(int j=0;j<tempZ.length;j++){
			node[0][j] = new svm_node();
			node[0][j].index = j+1;	node[0][j].value=tempZ[j];
		}
		double pp = svm.svm_predict(g, node[0]);
		double ans = (svm.svm_predict(g, node[0])+1)/2;
		return ans;
	}
	
	
	public Enumeration listOptions() {
	    
	    Vector<Option> tempVector = new Vector(2);
	    tempVector.addElement(new Option("\tmaxit:Number of max iterations \t (default 5)\n",
					    "M", 1, "-M <double>"));
	    tempVector.addElement(new Option("\tbeta: Starting penalty parameter \t (default 1)",
	    				"penalty", 1, "-penalty <double>"));
	    tempVector.addElement(new Option("\ttol: minimum relative decrease in error to keep iterating \t (default 1e-4)",
	    				"tol", 1, "-tol <double"));
	    tempVector.addElement(new Option("\tzLength: Set the length of Z dimensions to be reduced to",
				"L", 1, "-L <double>"));
	    tempVector.addElement(new Option("\tregType: Set which regressor to use, 0 = LinearRegression, 1 = RBFRegression \t (default 1)",
				"R", 1, "-R <double>"));
	    tempVector.addElement(new Option("\tregOptions: Options to be passed directly into the regressor"
    			+ "\t(default {-S 1  -C -minimal -R 0 do-not-check-capabilities} for LinearRegression, {-N 30 -C 1 -do-not-check-capabilities} for RBFRegressor",
				"O", 0, "-O <String delimited by spaces>"));
	    return tempVector.elements();
	}
	
	public void setOptions(String[] options) throws Exception {

	    String tempStr = Utils.getOption('M', options);
	    if (tempStr.length() != 0) {
	      setMaxIt(new Integer(tempStr).intValue() );
	    }
	    else{
	    	setMaxIt(5);
	    }
	    tempStr = Utils.getOption("penalty", options);
	    if(tempStr.length() !=0 ){
	    	setBeta(new Double(tempStr).doubleValue());
	    }
	    else{
	    	setBeta(1);
	    }
	    tempStr = Utils.getOption("tol", options);
	    if(tempStr.length() != 0){
	    	setTol(new Double(tempStr).doubleValue());
	    }
	    else{
	    	setTol(.00001);
	    }
	    tempStr = Utils.getOption('L', options);
	    if(tempStr.length() != 0){
	    	setL(new Integer(tempStr).intValue());
	    }
	    else{
	    	setL(2);
	    }
	    tempStr = Utils.getOption('R', options);
	    if(tempStr.length() != 0){
	    	setRegType(new Integer(tempStr).intValue());
	    }
	    else{
	    	setRegType(0);
	    }
	    tempStr = Utils.getOption('O', options);
	    if(tempStr.length() != 0){
	    	setRegOptions(tempStr, this.regType);
	    }
	    else{
	    	if(getRegType() == 0){
	    		setRegOptions("-S 1  -C -minimal -R 0 -do-not-check-capabilities", this.regType);
	    	}
	    	else{
	    		setRegOptions("-N 50 -C 1 -do-not-check-capabilities", this.regType);
	    	}
	    }
	}
	
	@Override
	public String[] getOptions() {
		List<String> options = new ArrayList<String>();
		
		options.add("-M");
		options.add("" + getMaxIt());
		
		options.add("-penalty");
		options.add("" + getBeta());
		
		options.add("-tol");
		options.add("" + getTol());
		
		options.add("-L");
		options.add("" + getL());
		
		options.add("-R");
		options.add("" + getRegType());
		
		if(getRegOptions().length() != 0){
			options.add("-O");
			options.add("" + getRegOptions());
		}
		return options.toArray(new String[options.size()]);
	}
	
	//Utility class ported in
	private double[][] fClassify(Instances X, Vector<AbstractClassifier> f) throws Exception{
		int col = f.size();
		double[][] results = new double[X.numInstances()][col];
		for(int i = 0; i < X.numInstances(); i++){
			DenseInstance xInst = (DenseInstance) X.get(i);
			for(int j = 0; j < col; j++){
				results[i][j] = f.get(j).classifyInstance(xInst);
			}
		}
		return results;
	}
	
	private double[] gClassify(svm_node[][] Z, svm_model g) throws Exception{
		
		int N = Z.length;
		double[] results = new double[N];
		for(int i = 0; i < N; i++){			
			results[i] = svm.svm_predict(g, Z[i]);
		}
		return results;
		
	}
	
	private AbstractClassifier getRegressor(int regType) throws Exception{
		switch(regType){
			case 0: LinearRegression tempLin = new LinearRegression();					
					return tempLin;
			case 1: RBFRegressor tempRBF = new RBFRegressor();					
					System.out.println("Using RBF");
					return tempRBF;
		}
		return null;
	}
	
	private double frob(Matrix m){
		
		double sum = 0;
		for(int i = 0; i < m.getRowDimension(); i++){
			for(int j = 0; j < m.getColumnDimension(); j++){
				sum+=Maths.square(m.get(i,j));
			}
		}
		return Math.sqrt(sum);
	}
		
	private double[][] ZOpt(double[][] Z, double[][] fX, double[] labels, double beta, svm_model g){
		//double bias = g.getBias();
		int N = Z.length; int L = Z[0].length;
		double c = (g.param.C)/beta;
		double b = -1*g.rho[0];
		int nSV = g.l;
		double[] w = new double[L]; 	double sw = 0;
		for(int l = 0; l<L;l++){
			w[l]=0;
			for(int i = 0; i <nSV; i++){				
				w[l] += g.sv_coef[0][i]*g.SV[i][l].value;				
			}
			sw+=(w[l])*(w[l]);
		}
		
		double[] LAMBDA=new double[N];
		for(int i = 0; i < N; i++){
			LAMBDA[i]=0;
			for(int j = 0; j < L ; j++){
				LAMBDA[i] +=  fX[i][j]*w[j];
			}
			LAMBDA[i] += b; LAMBDA[i] *= labels[i]; LAMBDA[i] = 1- LAMBDA[i] ;
			if(LAMBDA[i]>0)
			{
				LAMBDA[i] = (2*LAMBDA[i])/sw;
			}
			else
			{
				LAMBDA[i] =0;
			}
		}
		
		for(int i=0;i<N;i++)
		{
			double temp = (.5)*(Math.min(LAMBDA[i], c))*labels[i];
			for(int j = 0; j < L; j++)
			{
				Z[i][j] = (temp*(w[j])) + fX[i][j];				
			}
		}
		
		return Z;
	}
		
	private double gError(double[][] Z, double[] labels, svm_model g){
		//double bias = g.getBias();
		int N = Z.length; int L = Z[0].length;
		double c = (g.param.C);
		double b = -1*g.rho[0];
		int nSV = g.l;
		double[] w = new double[L]; 	double sw = 0;
		for(int l = 0; l<L;l++){
			w[l]=0;
			for(int i = 0; i <nSV; i++){				
				w[l] += g.sv_coef[0][i]*g.SV[i][l].value;				
			}
			sw+=(w[l])*(w[l]);
		}				
		
		double tSlacks; double sSlacks = 0;
		for(int i = 0; i < N; i++){
			tSlacks=0;
			for(int j = 0; j < L ; j++){
				tSlacks +=  Z[i][j]*w[j];
			}
			tSlacks += b; tSlacks *= labels[i]; tSlacks = 1- tSlacks ;
			if(tSlacks<0)
			{
				tSlacks =0;
			}
			sSlacks +=tSlacks;
		}
		double err = (0.5*sw)+ (c*sSlacks);
		return err;
		
	}
	
	private double fError(double[][] Z,double[][] fX){
		//double bias = g.getBias();
		int N = Z.length; int L = Z[0].length;
		double err = 0;
		for(int i =0; i < N; i++)
		{
			for(int j = 0; j < L; j++)
			{			
				err += Math.pow((Z[i][j] - fX[i][j]), 2)  ;  
			}
		}		
		return (err);
		
	}
	
	private double[][] matMult(double[][] A, double[][] B) {

	int aRows = A.length;
    int aColumns = A[0].length;    
    int bColumns = B[0].length;    

    double[][] C = new double[aRows][bColumns];    

    for (int i = 0; i < aRows; i++) { // aRow
        for (int j = 0; j < bColumns; j++) { // bColumn
        		C[i][j] = 0.00000;
            for (int k = 0; k < aColumns; k++) { // aColumn
                C[i][j] += A[i][k] * B[k][j];
            }
        }	
    }

    return C;
	}
	
	public static void main(String[] args) throws IOException {
		//String[] argv = {"-t", "/home/ramin/spiralsTrain.arff", "-T","/home/ramin/spiralsTest.arff", "-O","-N 50","-M","2","-R", "1", "-L", "2","-penalty",".01"};
	  //String[] argv = {"-t", "/home/ramin/lindataTrain.arff", "-T","/home/ramin/lindataTest.arff", "-M","5","-R", "0", "-L", "1","-p","2"};
		//String[] argv = {"-t", "/home/ramin/spiralsTrain.arff", "-T","/home/ramin/spiralsTest.arff", "-M","5","-R", "0", "-L", "1","-p","2"};
		//String[] argv = {"-t", "/home/ramin/spirals1.arff", "-T","/home/ramin/spirals1.arff","-R", "0", "-L", "2","-penalty","2","-tol",".01"};
	  //runClassifier(new ldsvm(), argv);
		
		 runClassifier(new ldsvm(), args);
	}


}
